# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import re
from datetime import datetime
from datetime import timedelta
from datetime import timezone


def hhmmss(seconds: float | None, threshold: float = 2.0) -> str:
    if seconds is None:
        return "--:--:--"
    t = datetime.fromtimestamp(seconds)
    utc = datetime.fromtimestamp(seconds, timezone.utc)
    if seconds < threshold:
        return datetime.strftime(utc, "%H:%M:%S.%f")[:-4]
    return datetime.strftime(utc, "%H:%M:%S")


def time_in_seconds(arg: int | float | str) -> float:
    if isinstance(arg, (float, int)):
        return float(arg)
    duration = Duration.from_str(arg)
    return duration.total_seconds()


class Duration:
    """Support for GO lang's duration format"""

    _nanosecond_size = 1
    _microsecond_size = 1000 * _nanosecond_size
    _millisecond_size = 1000 * _microsecond_size
    _second_size = 1000 * _millisecond_size
    _minute_size = 60 * _second_size
    _hour_size = 60 * _minute_size
    _day_size = 24 * _hour_size
    _week_size = 7 * _day_size
    _month_size = 30 * _day_size
    _year_size = 365 * _day_size

    units = {
        "ns": _nanosecond_size,
        "us": _microsecond_size,
        "µs": _microsecond_size,
        "μs": _microsecond_size,
        "ms": _millisecond_size,
        "s": _second_size,
        "m": _minute_size,
        "h": _hour_size,
        "d": _day_size,
        "w": _week_size,
        "mm": _month_size,
        "y": _year_size,
    }

    _re = re.compile(r"([\d\.]+)([a-zµμ]+)")

    @staticmethod
    def from_str(duration: str) -> timedelta:
        """Parse a duration string to a datetime.timedelta"""

        original = duration

        if not duration:
            return timedelta()
        elif duration in ("0", "+0", "-0"):
            return timedelta()

        sign = 1
        if duration[0] in "+-":
            sign = -1 if duration[0] == "-" else 1
            duration = duration[1:]

        if re.search(r"(?xm)(?:\s|^)([-+]*(?:\d+\.\d*|\.?\d+)(?:[eE][-+]?\d+)?)(?=\s|$)", duration):
            return timedelta(seconds=sign * float(duration))
        elif re.search(r"^\d{1,2}:\d{1,2}:\d{1,2}(\.\d+)?$", duration):
            hours, minutes, seconds = [float(_) for _ in duration.split(":")]
            units = Duration.units
            microseconds = hours * units["h"] + minutes * units["m"] + seconds * units["s"]
            print(microseconds / Duration._microsecond_size)
            return timedelta(microseconds=sign * microseconds / Duration._microsecond_size)
        elif re.search(r"^\d{1,2}:\d{1,2}(\.\d+)?$", duration):
            minutes, seconds = [float(_) for _ in duration.split(":")]
            microseconds = minutes * 60.0 + seconds * 1.0
            return timedelta(microseconds=sign * microseconds / Duration._microsecond_size)

        matches = list(Duration._re.finditer(duration))
        if not matches:
            raise DurationError("Invalid duration {}".format(original))
        if matches[0].start() != 0 or matches[-1].end() != len(duration):
            raise DurationError("Extra chars at start or end of duration {}".format(original))

        total = 0.0
        for match in matches:
            value, unit = match.groups()
            if unit not in Duration.units:
                raise DurationError("Unknown unit {} in duration {}".format(unit, original))
            try:
                total += float(value) * Duration.units[unit]
            except Exception:
                raise DurationError("Invalid value {} in duration {}".format(value, original))

        microseconds = total / Duration._microsecond_size
        return timedelta(microseconds=sign * microseconds)

    @staticmethod
    def to_str(delta: timedelta, extended: bool = False) -> str:
        """Format a datetime.timedelta to a duration string"""

        total_seconds = delta.total_seconds()
        sign = "-" if total_seconds < 0 else ""
        nanoseconds = abs(total_seconds * Duration._second_size)

        if abs(total_seconds) < 1:
            result_str = Duration._to_str_small(nanoseconds, extended)
        else:
            result_str = Duration._to_str_large(nanoseconds, extended)

        return "{}{}".format(sign, result_str)

    @staticmethod
    def _to_str_small(nanoseconds: float, extended: bool) -> str:
        result_str = ""

        if not nanoseconds:
            return "0"

        milliseconds = int(nanoseconds / Duration._millisecond_size)
        if milliseconds:
            nanoseconds -= Duration._millisecond_size * milliseconds
            result_str += "{:g}ms".format(milliseconds)

        microseconds = int(nanoseconds / Duration._microsecond_size)
        if microseconds:
            nanoseconds -= Duration._microsecond_size * microseconds
            result_str += "{:g}us".format(microseconds)

        if nanoseconds:
            result_str += "{:g}ns".format(nanoseconds)

        return result_str

    @staticmethod
    def _to_str_large(nanoseconds: float, extended: bool) -> str:
        result_str = ""

        if extended:
            years = int(nanoseconds / Duration._year_size)
            if years:
                nanoseconds -= Duration._year_size * years
                result_str += "{:g}y".format(years)

            months = int(nanoseconds / Duration._month_size)
            if months:
                nanoseconds -= Duration._month_size * months
                result_str += "{:g}mm".format(months)

            days = int(nanoseconds / Duration._day_size)
            if days:
                nanoseconds -= Duration._day_size * days
                result_str += "{:g}d".format(days)

        hours = int(nanoseconds / Duration._hour_size)
        if hours:
            nanoseconds -= Duration._hour_size * hours
            result_str += "{:g}h".format(hours)

        minutes = int(nanoseconds / Duration._minute_size)
        if minutes:
            nanoseconds -= Duration._minute_size * minutes
            result_str += "{:g}m".format(minutes)

        seconds = float(nanoseconds) / float(Duration._second_size)
        if seconds:
            nanoseconds -= Duration._second_size * seconds
            result_str += "{:g}s".format(seconds)

        return result_str


class DurationError(ValueError):
    """duration error"""



import io
import os
import shutil

from ..config import Config
from ..hookspec import hookimpl
from ..submit import slurm
from .base import HPCLauncher
from .base import LaunchSpecs


class SrunLauncher(HPCLauncher):
    def __init__(self, config: Config | None = None) -> None:
        super().__init__(config=config)
        if not self.exec.endswith("srun"):
            raise ValueError("SrunLauncher: expected exec = srun")
        if self.config.get("machine:resources") is None:
            if sinfo := slurm.read_sinfo():
                self.config.set("machine:resources", [sinfo])

    @staticmethod
    def matches(arg: str) -> bool:
        return os.path.basename(arg) == "srun"

    def join_specs(
        self,
        launchspecs: "LaunchSpecs",
        local_options: list[str] | None = None,
        global_options: list[str] | None = None,
        pre_options: list[str] | None = None,
    ) -> list[str]:
        """Count the total number of processes and write a srun.conf file to
        split the jobs across ranks

        """
        if len(launchspecs) <= 1:
            return super().join_specs(
                launchspecs, local_options=local_options, global_options=global_options
            )

        local_options = list(local_options or [])
        local_options.extend(self.config.get("launch:local_options"))
        global_options = list(global_options or [])
        global_options.extend(self.config.get("launch:default_options"))
        pre_options = list(pre_options or [])
        pre_options.extend(self.config.get("launch:pre_options"))

        np: int = 0
        fp = io.StringIO()
        for p, spec in launchspecs:
            ranks: str
            if p is not None:
                ranks = f"{np}-{np + p - 1}"
                np += p
            else:
                ranks = str(np)
                np += 1
            i = self.argp(spec)
            fp.write(ranks)
            for opt in local_options:
                fp.write(f" {self.expand(opt, np=np)}")
            for opt in pre_options:
                fp.write(f" {self.expand(opt, np=np)}")
            for arg in spec[i:]:
                fp.write(f" {self.expand(arg, np=p)}")
            fp.write("\n")
        file = "launch-multi-prog.conf"
        with open(file, "w") as fh:
            fh.write(fp.getvalue())
        cmd = [os.fsdecode(self.exec)]
        required_resources = self.config.compute_required_resources(ranks=np)
        for opt in global_options:
            cmd.append(self.expand(opt, **required_resources))
        cmd.extend([f"-n{np}", "--multi-prog", file])
        return cmd

    @staticmethod
    def argp(args: list[str]) -> int:
        for i, arg in enumerate(args):
            if shutil.which(arg):
                return i
        return -1


@hookimpl
def hpc_connect_launcher(config: Config) -> HPCLauncher | None:
    if SrunLauncher.matches(config.get("launch:exec")):
        return SrunLauncher(config=config)
    return None



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import pluggy


class HPCConnectPluginManager(pluggy.PluginManager):
    def __init__(self):
        from . import hookspec
        from . import launch
        from . import submit

        super().__init__("hpc_connect")
        self.add_hookspecs(hookspec)
        for module in submit.plugins:
            self.register(module)
        for module in launch.plugins:
            self.register(module)
        self.load_setuptools_entrypoints("hpc_connect")



import subprocess
from typing import Any

from ..config import Config
from . import base
from . import srun
from .base import HPCLauncher

plugins = (base, srun)


def factory(config: Config | None = None) -> HPCLauncher:
    config = config or Config()
    launcher = config.pluginmanager.hook.hpc_connect_launcher(config=config)
    if launcher is None:
        exec = config.get("launch:exec")
        raise ValueError(f"No matching launcher manager for {exec!r}")
    return launcher


def launch(args: list[str], **kwargs: Any) -> subprocess.CompletedProcess:
    launcher = factory()
    return launcher(args, **kwargs)



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import os

import pytest


@pytest.fixture(scope="session", autouse=True)
def add_mock_path():
    save_env = os.environ.copy()
    mock = os.path.join(os.path.dirname(__file__), "mock")
    assert os.path.exists(mock), mock
    os.environ["PATH"] = f"{mock}:{os.environ['PATH']}"
    yield
    os.environ.clear()
    os.environ.update(save_env)



#!/usr/bin/env python
import argparse
import glob
import logging
import os
import shutil
import subprocess
import sys
from contextlib import contextmanager
from typing import Generator

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")
    p = subparsers.add_parser("build", help="Build wheels")
    p.add_argument("-l", action="store_true", default=False, help="Keep local")
    p = subparsers.add_parser("deploy", help="Upload to pypi index")
    p.add_argument("-u", help="twine username")
    p.add_argument("-p", help="twine password")
    args = parser.parse_args()
    if args.command == "build":
        return build_and_test(local=args.l)
    elif args.command == "deploy":
        return deploy(twine_username=args.u, twine_password=args.p)
    raise ValueError(f"Unknown command {args.command}")


def build_and_test(local: bool = False) -> None:
    build(local=local)
    test()


def build(local: bool = False) -> None:
    logger.info("Building wheel")
    with branch("production"):
        if os.path.exists("dist"):
            shutil.rmtree("dist")
        merge("main", local=local)
        if not local:
            subprocess.run(["git", "push", "origin", "production"], check=True)
        with virtual_env() as env:
            subprocess.run([env.bin.python3, "-m", "pip", "install", "build"], check=True)
            subprocess.run([env.bin.python3, "-m", "build", "--wheel"], check=True)
    logger.info("Done building wheels")
    return


def test() -> None:
    logger.info("Testing")
    assert os.path.exists("./dist"), "build wheels first"
    with virtual_env() as env:
        dist_dir = os.path.abspath("./dist")
        url = f"file://{dist_dir}"
        subprocess.run(
            [env.bin.python3, "-m", "pip", "install", "--find-links", url, "hpc-connect"], check=True
        )
        subprocess.run([env.bin.python3, "-m", "hpc_connect", "-h"], check=True)
    logger.info("Done testing")


def deploy(*, twine_username: str | None = None, twine_password: str | None = None) -> None:
    assert os.path.exists("./dist"), "build wheels first"
    logger.info("Deploying")
    with virtual_env() as env:
        if twine_username is not None:
            os.environ["TWINE_USERNAME"] = twine_username
        if "TWINE_USERNAME" not in os.environ:
            raise MissingEnvironmentVariableError("TWINE_USERNAME")
        if twine_password is not None:
            os.environ["TWINE_PASSWORD"] = twine_password
        if "TWINE_PASSWORD" not in os.environ:
            raise MissingEnvironmentVariableError("TWINE_PASSWORD")
        subprocess.run([env.bin.python3, "-m", "pip", "install", "twine"], check=True)
        wheels = glob.glob("./dist/*.whl")
        cmd = [
            env.bin.python3,
            "-m",
            "twine",
            "upload",
            "--skip-existing",
        ]
        cmd.extend(wheels)
        subprocess.run(cmd, check=True)
    logger.info("Done deploying")


class Prefix(str):
    def __getattr__(self, name: str) -> "Prefix":
        return Prefix(os.path.join(self, name))


@contextmanager
def virtual_env(preserve: bool = False) -> Generator[Prefix, None, None]:
    save_env = os.environ.copy()
    try:
        if os.path.exists("venv"):
            shutil.rmtree("venv")
        subprocess.run([sys.executable, "-m", "venv", "venv", "--system-site-packages"], check=True)
        save_env.pop("VIRTUAL_ENV", None)
        save_env.pop("VIRTUAL_ENV_PROMPT", None)
        save_env.pop("PYTHONHOME", None)
        venv = os.path.abspath("./venv")
        os.environ["VIRTUAL_ENV"] = venv
        os.environ["PATH"] = f"{venv}/bin:{os.environ['PATH']}"
        yield Prefix(venv)
    finally:
        os.environ.clear()
        os.environ.update(save_env)
        if not preserve:
            shutil.rmtree("venv")


@contextmanager
def branch(name: str) -> None:
    cwb = subprocess.check_output(["git", "branch", "--show-current"], encoding="utf-8").strip()
    try:
        subprocess.run(["git", "checkout", name], check=True)
        yield
    finally:
        subprocess.run(["git", "checkout", cwb], check=True)


def merge(name, local: bool = False):
    cwb = subprocess.check_output(["git", "branch", "--show-current"], encoding="utf-8").strip()
    git = lambda *args: subprocess.run(["git", *args], check=True)
    if not local:
        git("reset", "--hard", "HEAD")
        git("checkout", name)
        git("pull", "origin", "main")
        git("checkout", cwb)
    git("merge", "--no-ff", name, "-m", f"Merge remote-traching branch {name!r} into {cwb}")


class MissingEnvironmentVariableError(Exception):
    pass


if __name__ == "__main__":
    sys.exit(main())



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import importlib.resources

import jinja2

from .time import hhmmss


def make_template_env(*dirs: str) -> jinja2.Environment:
    """Returns a configured environment for template rendering."""
    template_dirs: set[str] = {str(importlib.resources.files("hpc_connect").joinpath("templates"))}
    template_dirs.update(dirs)
    loader = jinja2.FileSystemLoader(tuple(template_dirs))
    env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
    env.globals["hhmmss"] = hhmmss
    return env



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import io
import os
import tempfile
from contextlib import contextmanager

import hpc_connect


@contextmanager
def tmp_environ():
    save_env = os.environ.copy()
    try:
        os.environ["SLURM_NTASKS_PER_NODE"] = "10"
        os.environ["SLURM_NNODES"] = "1"
        os.environ["SLURM_GPUS"] = "0"
        yield
    finally:
        os.environ.clear()
        os.environ.update(save_env)


def test_basic():
    backend = hpc_connect.submit.slurm.SlurmSubmissionManager()
    with io.StringIO() as fh:
        backend.write_submission_script(
            "my-job",
            ["ls"],
            fh,
            cpus=1,
            nodes=1,
            output="my-out.txt",
            error="my-err.txt",
            qtime=1.0,
            variables={"MY_VAR": "SPAM"},
        )
        text = fh.getvalue()
    assert "#!/bin/sh" in text
    assert "#SBATCH --nodes=1" in text
    assert "#SBATCH --time=00:00:01" in text
    assert "#SBATCH --job-name=my-job" in text
    assert "#SBATCH --error=my-err.txt" in text
    assert "#SBATCH --output=my-out.txt" in text
    assert "export MY_VAR=SPAM" in text
    assert "printenv || true" in text
    assert "ls" in text


def test_parse_script_args():
    with tempfile.NamedTemporaryFile("w") as fh:
        fh.write("""\
#!/bin/sh
#SBATCH --nodes=1
#SBATCH --time=00:00:01
#SBATCH --job-name=my-job
#SBATCH --error=my-err.txt
#SBATCH --output=my-out.txt
#SBATCH --clusters=flight,eclipse
export MY_VAR=SPAM
printenv || true
ls""")
        fh.seek(0)
        ns = hpc_connect.submit.slurm.SlurmProcess.parse_script_args(fh.name)
        assert ns.clusters == "flight,eclipse"



"""
Overview
--------

`hpcc` is a lightweight and configurable wrapper around HPC schedulers and program launchers like
`mpiexec` or `srun`. `hpcc` provides a single interface to multiple backends, simplifying the
process of running jobs in an HPC environment.  backend implementation.

Configuration
-------------

The default behavior of `hpcc` can be changed by providing a yaml configuration file.  The default
configuration is:

.. code-block:: yaml

   hpc_connect:
     config:
       debug: false
     launch:
       exec: mpiexec  # the launch backend.
       numproc_flag: -n  # Flag to pass to the backend before giving it the number of processors to run on.
       local_options: []  # Options to pass to the backend before any other arguments.
       default_options: []  # Options to pass to the backend before any other arguments.
       pre_options: []  # Command line options placed immediately before program to run
       mappings: {-n: <numproc_flag>}  # Mapping of flag provided on the command line to flag passed to the backend
    submit:
      backend: null
      default_options: []

Configurations are read from:

1. Local configuration: ./hpc_connect.yaml
2. Global configuration [1]: ~/.config/hpc_connect/config.yaml
3. Site configuration [2]: sys.prefix/etc/hpc_connect/config.yaml

[1] The global configuration will be read from the HPC_CONNECT_GLOBAL_CONFIG environment variable, if set
[2] The site configuration will be read from the HPC_CONNECT_SITE_CONFIG environment variable, if set

Configuration settings can also be modified through the following environment variables:

* HPCC_LAUNCH_EXEC
* HPCC_LAUNCH_NUMPROC_FLAG
* HPCC_LAUNCH_LOCAL_OPTIONS
* HPCC_LAUNCH_DEFAULT_OPTIONS
* HPCC_LAUNCH_PRE_OPTIONS
* HPCC_LAUNCH_MAPPINGS
* HPCC_SUBMIT_BACKEND
* HPCC_SUBMIT_DEFAULT_OPTIONS

"""

import argparse
import sys
from types import ModuleType

from ..config import Config
from . import config
from . import launch
from . import submit

_commands: dict[str, ModuleType] = {}


def main(argv: list[str] | None = None) -> int:
    parser = make_parser()
    args, extra_args = parser.parse_known_args(argv or sys.argv[1:])
    if args.info:
        print(__doc__)
        return 0
    args.extra_args = extra_args

    module = _commands[args.command]
    cfg = Config()
    cfg.set_main_options(args)
    module.execute(cfg, args)
    return 0


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--info", action="store_true", help="Show additional information and exit.")
    parser.add_argument(
        "-c",
        dest="config_mods",
        action="append",
        metavar="path",
        help="colon-separated path to config that should be set, e.g. 'config:default:true'",
    )
    subparsers = parser.add_subparsers(dest="command")
    add_command(subparsers, config)
    add_command(subparsers, launch)
    add_command(subparsers, submit)
    return parser


def add_command(subparsers: argparse._SubParsersAction, module: ModuleType) -> None:
    name = module.__name__.split(".")[-1].lower()
    parser = subparsers.add_parser(
        name, add_help=getattr(module, "add_help", True), help=module.description
    )
    module.setup_parser(parser)
    _commands[name] = module



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
from typing import TYPE_CHECKING

import pluggy

if TYPE_CHECKING:
    from .config import Config as HPCConfig
    from .launch import HPCLauncher
    from .submit import HPCSubmissionManager


hookspec = pluggy.HookspecMarker("hpc_connect")
hookimpl = pluggy.HookimplMarker("hpc_connect")


@hookspec(firstresult=True)
def hpc_connect_submission_manager(config: "HPCConfig") -> "HPCSubmissionManager | None":
    """HPC scheduler implementation"""


@hookspec(firstresult=True)
def hpc_connect_launcher(config: "HPCConfig") -> "HPCLauncher | None":
    """HPC scheduler implementation"""



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import importlib.resources
import logging
import math
import multiprocessing
import multiprocessing.synchronize
import os
import subprocess
import time
from concurrent.futures import CancelledError
from datetime import timedelta
from typing import Any

import flux  # type: ignore
from flux import Flux  # type: ignore
from flux.job import FluxExecutor  # type: ignore
from flux.job import FluxExecutorFuture  # type: ignore
from flux.job import Jobspec  # type: ignore
from flux.job import JobspecV1  # type: ignore

from ..config import Config
from ..util import time_in_seconds
from .base import HPCProcess
from .base import HPCSubmissionFailedError
from .base import HPCSubmissionManager

logger = logging.getLogger("hpc_connect")


class FluxProcess(HPCProcess):
    JOB_TIMEOUT_CODE = 66

    def __init__(self, name: str, future: FluxExecutorFuture) -> None:
        self.fh = Flux()
        self.name = name
        self.fut: FluxExecutorFuture = future
        self._jobid: int | None = None
        self._rc: int | None = None

        def set_returncode(fut: FluxExecutorFuture):
            try:
                info = flux.job.result(self.fh, fut.jobid())
                self.returncode = info.returncode
            except (CancelledError, Exception):
                self.returncode = 1

        def set_jobid(fut: FluxExecutorFuture):
            try:
                self.jobid = fut.jobid()
                logger.debug(f"submitted job {self.jobid} for {self.name}")
            except CancelledError:
                self.returncode = 1
            except Exception as e:
                raise HPCSubmissionFailedError(e)

        self.fut.add_jobid_callback(set_jobid)
        self.fut.add_done_callback(set_returncode)

    @property
    def jobid(self) -> int | None:
        return self._jobid

    @jobid.setter
    def jobid(self, arg: int) -> None:
        self._jobid = arg

    @property
    def returncode(self) -> int | None:
        return self._rc

    @returncode.setter
    def returncode(self, arg: int) -> None:
        self._rc = arg

    def poll(self) -> int | None:
        return self.returncode

    def cancel(self) -> None:
        logger.warning(f"Canceling flux job {self.jobid}")
        try:
            flux.job.cancel(self.fh, self.jobid)
        except OSError:
            logger.debug(f"Job {self.jobid} is inactive, cannot cancel")
        except Exception:
            logger.error(f"Failed to cancel job {self.jobid}")
        self.returncode = 1


class FluxMultiProcess(HPCProcess):
    def __init__(
        self,
        lock: multiprocessing.synchronize.RLock,
        procs: list[FluxProcess] | None = None,
    ) -> None:
        self.lock = lock
        self.procs = procs or []

    @property
    def returncode(self) -> int | None:
        rcs = [p.returncode for p in self.procs if p is not None]
        if not rcs:
            return None
        return max(rcs)  # type: ignore

    @returncode.setter
    def returncode(self, arg: int) -> None:
        raise NotImplementedError

    def append(self, proc: FluxProcess) -> None:
        self.procs.append(proc)

    def pop(self, /, i: int = -1) -> FluxProcess:
        return self.procs.pop(i)

    def cancel(self) -> None:
        with self.lock:
            for proc in self.procs:
                proc.cancel()

    def poll(self) -> int | None:
        stat: list[int | None] = []
        for proc in self.procs:
            stat.append(proc.poll())
        if any([_ is None for _ in stat]):
            return None
        return max(stat)  # type: ignore


def parse_resource_info(output: str) -> dict[str, int] | None:
    """Parses the output from `flux resource info` and returns a dictionary of resource values.

    The expected output format is "1 Nodes, 32 Cores, 1 GPUs".

    Returns:
        dict: A dictionary containing the resource values with the following keys:
            - nodes (int): The number of nodes.
            - cpu (int): The number of CPU cores.
            - gpu (int): The number of GPU devices.
    """
    parts = output.split(", ")
    vals = [int(p.split()[0]) for p in parts]
    if len(vals) != 3:
        return None
    return {"nodes": vals[0], "cpu": vals[1], "gpu": vals[2]}


def read_resource_info() -> dict[str, Any] | None:
    try:
        output = subprocess.check_output(["flux", "resource", "info"], encoding="utf-8")
    except subprocess.CalledProcessError:
        return None
    if totals := parse_resource_info(output):
        # assume homogenous resources
        nodes = totals["nodes"]
        info: dict = {
            "type": "node",
            "count": nodes,
            "resources": [
                {
                    "type": "socket",
                    "count": 1,
                    "resources": [
                        {
                            "type": "cpu",
                            "count": int(totals["cpu"] / nodes),
                        },
                        {
                            "type": "gpu",
                            "count": int(totals["gpu"] / nodes),
                        },
                    ],
                }
            ],
        }
        return info
    return None


class FluxSubmissionManager(HPCSubmissionManager):
    """Setup and submit jobs to the Flux scheduler"""

    name = "flux"
    lock: multiprocessing.synchronize.RLock = multiprocessing.RLock()

    def __init__(self, config: Config | None = None) -> None:
        super().__init__(config=config)
        self.flux: FluxExecutor | None = FluxExecutor()
        self.fh = Flux()
        if self.config.get("machine:resources") is None:
            if info := read_resource_info():
                self.config.set("machine:resources", [info], scope="defaults")

    @property
    def supports_subscheduling(self) -> bool:
        return True

    @staticmethod
    def matches(name: str | None) -> bool:
        return name is not None and name.lower() == "flux"

    @property
    def submission_template(self) -> str:
        if "HPCC_FLUX_SUBMIT_TEMPLATE" in os.environ:
            return os.environ["HPCC_FLUX_SUBMIT_TEMPLATE"]
        return str(importlib.resources.files("hpc_connect").joinpath("templates/flux.sh.in"))

    def create_jobspec(
        self,
        name: str,
        script: str,
        duration: float | timedelta | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        #
        cpus: int | None = None,
        gpus: int | None = None,
        nodes: int | None = None,
        exclusive: bool = True,
    ) -> Jobspec:
        """
        Create the flux jobspec for the executable script ``script``.  The number of tasks
        ``tasks`` is the number of MPI tasks that ``script`` launches, not the number of
        copies of ``script`` that flux should launch.

        """
        kwds: dict[str, Any] = {"command": [script], "exclusive": exclusive}
        alloc = self.get_alloc_settings(cpus, gpus, nodes)
        kwds.update(alloc)
        jobspec = JobspecV1.from_nest_command(**kwds)
        jobspec.setattr("system.job.name", name)
        jobspec.stdout = output or "job-ouput.txt"
        jobspec.stderr = error or output or "job-error.txt"
        jobspec.duration = duration or 60.0  # duration is in seconds
        env = os.environ.copy()
        if variables:
            for key, val in variables.items():
                if val is None:
                    env.pop(key, None)
                else:
                    env[key] = val
        jobspec.environment = env
        return jobspec

    def get_alloc_settings(
        self,
        cpus: int | None = None,
        gpus: int | None = None,
        nodes: int | None = None,
    ) -> dict[str, Any]:
        alloc: dict[str, Any] = {}
        if nodes is not None:
            if cpus is None:
                cpus = nodes * self.config.cpus_per_node
            if gpus is None:
                gpus = nodes * self.config.gpus_per_node
        else:
            cpus = cpus or 1
            gpus = gpus or 0
            nodes = self.config.nodes_required(max_cpus=cpus, max_gpus=gpus)

        alloc["num_nodes"] = nodes
        alloc["num_slots"] = nodes
        if nodes > 1:
            cpus = max(1, math.ceil(cpus / nodes))
            gpus = max(0, math.ceil(gpus / nodes))

        alloc["cores_per_slot"] = cpus
        alloc["gpus_per_slot"] = gpus
        return alloc

    def shutdown(self):
        with self.lock:
            if self.flux:
                time.sleep(0.25)
                if flux.job:
                    jobs = flux.job.JobList(self.fh, filters=["running", "pending"]).jobs()
                    if jobs:
                        logger.warning(f"{len(jobs)} active jobs remain after shutdown")
                self.flux.shutdown(wait=False, cancel_futures=True)
                self.flux = None

    def submit(
        self,
        name: str,
        args: list[str],
        scriptname: str | None = None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
        exclusive: bool = True,
        **kwargs: Any,
    ) -> FluxProcess:
        cpus = cpus or kwargs.get("tasks")
        duration = timedelta(seconds=time_limit_in_seconds(qtime, pad=60))
        script = self.write_submission_script(
            name,
            args,
            scriptname,
            qtime=max(1, duration.total_seconds() / 60.0),  # time limit in minutes in the script
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        assert script is not None
        jobspec = self.create_jobspec(
            name,
            script,
            duration=duration,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
            exclusive=exclusive,
        )
        fut = self.flux.submit(jobspec)  # type: ignore
        return FluxProcess(name, future=fut)

    def submitn(
        self,
        name: list[str],
        args: list[list[str]],
        scriptname: list[str] | None = None,
        qtime: list[float] | None = None,
        submit_flags: list[list[str]] | None = None,
        variables: list[dict[str, str | None]] | None = None,
        output: list[str] | None = None,
        error: list[str] | None = None,
        nodes: list[int] | None = None,
        cpus: list[int] | None = None,
        gpus: list[int] | None = None,
        **kwargs: Any,
    ) -> FluxMultiProcess:
        cpus = cpus or kwargs.get("tasks")  # backward compatible
        assert len(name) == len(args)
        procs = FluxMultiProcess(self.lock)
        submission_delay: float = 0.0
        if t := os.getenv("HPC_CONNECT_FLUX_SUBMITN_DELAY"):
            submission_delay = time_in_seconds(t)
        with self.lock:
            for i in range(len(name)):
                proc = self.submit(
                    name[i],
                    args[i],
                    select(scriptname, i, f"flux-submit-{i}"),
                    qtime=select(qtime, i),
                    submit_flags=select(submit_flags, i),
                    variables=select(variables, i),
                    output=select(output, i, f"flux-out-{i}"),
                    error=select(error, i, f"flux-err-{i}"),
                    nodes=select(nodes, i),
                    cpus=select(cpus, i),
                    gpus=select(gpus, i),
                    exclusive=False,
                )
                if submission_delay:
                    time.sleep(submission_delay)
                procs.append(proc)
        return procs

    def format_submission_data(
        self,
        name: str,
        args: list[str],
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        #
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
    ) -> dict[str, Any]:
        data = super().format_submission_data(
            name,
            args,
            qtime=qtime,
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        alloc = self.get_alloc_settings(cpus=cpus, gpus=gpus, nodes=nodes)
        data.update(alloc)
        return data


def select(arg: Any, i: int, default: Any = None) -> Any:
    if isinstance(arg, list):
        return arg[i]
    return arg or default


def time_limit_in_seconds(qtime: float | None, pad: int = 0) -> int:
    """Return the time limit in seconds. Guarenteed return value >= 1."""
    limit = math.ceil(qtime or 1)
    if pad > 0:
        limit += pad
    return limit



import os

import hpc_connect.config


def test_config_launch_basic(tmpdir):
    try:
        cwd = os.getcwd()
        os.chdir(tmpdir.strpath)

        config = hpc_connect.config.Config()
        config.set("launch:exec", "my-mpiexec")
        assert config.get("launch:exec") == "my-mpiexec"

        config.set("launch:local_options", ["-a", "-b"])
        assert config.get("launch:local_options") == ["-a", "-b"]
        config.add('launch:local_options:["-c", "-d"]')
        assert config.get("launch:local_options") == ["-a", "-b", "-c", "-d"]
        config.add('launch:local_options:-e')
        assert config.get("launch:local_options") == ["-a", "-b", "-c", "-d", "-e"]

        config.set("launch:mappings", {"-a": "-b", "-c": "-d"})
        assert config.get("launch:mappings") == {"-a": "-b", "-c": "-d"}
        config.add('launch:mappings:{"-e": "-f"}')
        assert config.get("launch:mappings") == {"-a": "-b", "-c": "-d", "-e": "-f"}
    finally:
        os.chdir(cwd)



import logging
import os
import shlex
import shutil
import subprocess
from typing import Any
from typing import Generator
from typing import Sequence

from ..config import Config
from ..hookspec import hookimpl


class HPCLauncher:
    def __init__(self, config: Config | None = None) -> None:
        self.config = config or Config()

    def __call__(
        self, args: list[str], echo: bool = False, **kwargs: Any
    ) -> subprocess.CompletedProcess:
        cmd = self.prepare_command_line(args)
        if echo:
            logging.getLogger(__name__).log(100, f"Command line: {shlex.join(cmd)}")
        return subprocess.run(cmd, **kwargs)

    @staticmethod
    def matches(arg: str) -> bool:
        return True

    def get_from_config(self, key: str, default: Any | None = None) -> Any:
        myexec = self.exec
        if value := self.config.get(f"launch:{myexec}:{key}"):
            return value
        elif value := self.config.get(f"launch:{os.path.basename(myexec)}:{key}"):
            return value
        else:
            value = self.config.get(f"launch:{key}")
            return default if value is None else value

    def set_config_var(self, key: str, value: Any, scope: str | None = None) -> None:
        myexec = self.exec
        self.config.set(f"launch:{os.path.basename(myexec)}:{key}", value, scope=scope)

    @property
    def exec(self) -> str:
        return self.config.get("launch:exec")

    @property
    def mappings(self) -> dict[str, str]:
        return self.get_from_config("mappings") or {}

    @property
    def numproc_flag(self) -> str:
        return self.get_from_config("numproc_flag", default="-n")

    @property
    def local_options(self) -> list[str]:
        return self.get_from_config("local_options") or []

    @property
    def global_options(self) -> list[str]:
        return self.get_from_config("default_options") or []

    @property
    def pre_options(self) -> list[str]:
        return self.get_from_config("pre_options") or []

    def prepare_command_line(self, args: list[str]) -> list[str]:
        parser = ArgumentParser(mappings=self.mappings, numproc_flag=self.numproc_flag)
        launchspecs = parser.parse_args(args)
        return self.join_specs(launchspecs)

    def join_specs(
        self,
        launchspecs: "LaunchSpecs",
        local_options: list[str] | None = None,
        global_options: list[str] | None = None,
        pre_options: list[str] | None = None,
    ) -> list[str]:
        local_options = list(local_options or [])
        local_options.extend(self.local_options)
        global_options = list(global_options or [])
        global_options.extend(self.global_options)
        pre_options = list(pre_options or [])
        pre_options.extend(self.pre_options)

        cmd = [os.fsdecode(self.exec)]
        np = sum(p for p in launchspecs.processes if p)
        required_resources = self.config.compute_required_resources(ranks=np)
        for opt in global_options:
            cmd.append(self.expand(opt, **required_resources))
        for p, spec in launchspecs:
            for opt in local_options:
                cmd.append(self.expand(opt, np=p))
            i = argp(spec)
            if i == -1:
                launch_opts, program_opts = [], spec
            else:
                launch_opts, program_opts = spec[:i], spec[i:]
            for opt in launch_opts:
                cmd.append(self.expand(opt, np=p))
            for opt in pre_options:
                cmd.append(self.expand(opt, np=p))
            for opt in program_opts:
                cmd.append(self.expand(opt, np=p))
            cmd.append(":")
        if cmd[-1] == ":":
            cmd.pop()
        return cmd

    @staticmethod
    def expand(arg: str, **kwargs: Any) -> str:
        return str(arg) % kwargs


class LaunchSpecs:
    def __init__(self) -> None:
        self.specs: list[list[str]] = []
        self.processes: list[int | None] = []

    def __len__(self) -> int:
        return len(self.specs)

    def __repr__(self) -> str:
        s = " : ".join(shlex.join(_) for _ in self.specs)
        return f"LaunchSpecs({s})"

    def __iter__(self) -> Generator[tuple[int | None, list[str]], None, None]:
        for i, spec in enumerate(self.specs):
            yield self.processes[i], spec

    def add(self, spec: list[str], processes: int | None) -> None:
        self.specs.append(list(spec))
        self.processes.append(processes)


class ArgumentParser:
    def __init__(self, *, mappings: dict[str, str] | None, numproc_flag: str = "-n") -> None:
        self.mappings: dict[str, str] = mappings or {}
        self.numproc_flag = numproc_flag
        if "-n" not in self.mappings:
            # always map -n to numproc_flag
            self.mappings["-n"] = self.numproc_flag

    def mapped(self, arg: str) -> str | None:
        if arg in self.mappings:
            return self.mappings[arg]
        # check for the case of long opt: arg=
        for pat, repl in self.mappings.items():
            if arg.startswith(f"{pat}="):
                return arg.replace(pat, repl)
        return None

    def parse_args(self, args: Sequence[str]) -> LaunchSpecs:
        """Inspect arguments to launch to infer number of processors requested"""
        launchspecs = LaunchSpecs()
        spec: list[str] = []
        processes: int | None = None
        command_seen: bool = False
        iter_args = iter(args or [])
        while True:
            try:
                arg = next(iter_args)
            except StopIteration:
                break
            if shutil.which(arg):
                command_seen = True
            if not command_seen:
                if new := self.mapped(arg):
                    if new.startswith("SUPPRESS="):
                        continue
                    elif new == "SUPPRESS":
                        next(iter_args)
                        continue
                    arg = new
                if arg == self.numproc_flag:
                    s = next(iter_args)
                    processes = int(s)
                    spec.extend([arg, s])
                elif arg.startswith(f"{self.numproc_flag}="):
                    i = len(self.numproc_flag) + 1
                    processes = int(arg[i:])
                    spec.append(arg)
                else:
                    spec.append(arg)
            elif arg == ":":
                # MPMD: end of this segment
                launchspecs.add(spec, processes)
                spec.clear()
                command_seen, processes = False, None
            else:
                spec.append(arg)

        if spec:
            launchspecs.add(spec, processes)

        return launchspecs


def argp(args: list[str]) -> int:
    for i, arg in enumerate(args):
        if shutil.which(arg):
            return i
    return -1


@hookimpl(trylast=True)
def hpc_connect_launcher(config: Config) -> HPCLauncher:
    return HPCLauncher(config=config)



import copy


def merge(dest, source):
    """Merges source into dest; entries in source take precedence over dest.

    This routine may modify dest and should be assigned to dest, in
    case dest was None to begin with, e.g.:

       dest = merge(dest, source)

    In the result, elements from lists from ``source`` will appear before
    elements of lists from ``dest``. Likewise, when iterating over keys
    or items in merged ``OrderedDict`` objects, keys from ``source`` will
    appear before keys from ``dest``.

    Config file authors can optionally end any attribute in a dict
    with `::` instead of `:`, and the key will override that of the
    parent instead of merging.
    """

    def they_are(t):
        return isinstance(dest, t) and isinstance(source, t)

    # If source is None, overwrite with source.
    if source is None:
        return None

    if they_are(list):
        dest[:] = dest + [x for x in source if x not in dest]
        return dest

    # Source dict is merged into dest.
    elif they_are(dict):
        # save dest keys to reinsert later -- this ensures that  source items
        # come *before* dest in OrderdDicts
        dest_keys = [dk for dk in dest.keys() if dk not in source]

        for sk, sv in source.items():
            # always remove the dest items. Python dicts do not overwrite
            # keys on insert, so this ensures that source keys are copied
            # into dest along with mark provenance (i.e., file/line info).
            merge_objects = sk in dest
            old_dest_value = dest.pop(sk, None)

            if merge_objects:
                dest[sk] = merge(old_dest_value, sv)
            else:
                # if sk ended with ::, or if it's new, completely override
                dest[sk] = copy.deepcopy(sv)

        # reinsert dest keys so they are last in the result
        for dk in dest_keys:
            dest[dk] = dest.pop(dk)

        return dest

    # If we reach here source and dest are either different types or are
    # not both lists or dicts: replace with source.
    return copy.copy(source)



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import argparse
import os
import shlex

from ..config import Config
from ..launch import factory as get_launcher

description = "Wrapper to launch tool, such as mpiexec or srun"
add_help = False


def setup_parser(parser: argparse.ArgumentParser) -> None:
    parser.usage = "%(prog)s [-h] [--help] ..."
    parser.add_argument(
        "-h",
        default=False,
        action="help",
        help="Show this message and exit.  For the backend's help, run 'hpcc launch --help'",
    )
    parser.add_argument(
        "--dryrun",
        action="store_true",
        default=False,
        help="Print command line to shell but do not execute",
    )


def execute(config: Config, args: argparse.Namespace) -> None:
    argv = list(args.extra_args)
    launcher = get_launcher(config)
    cmd = launcher.prepare_command_line(argv)
    if args.dryrun:
        print(shlex.join(cmd))
        return
    os.execvp(cmd[0], cmd)



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import sys

from . import command


def main(argv: list[str] | None = None) -> int:
    argv = argv or sys.argv[1:]
    return command.main(argv)


def launch(argv: list[str] | None = None) -> int:
    argv = argv or sys.argv[1:]
    argv.insert(0, "launch")
    return command.main(argv)


if __name__ == "__main__":
    sys.exit(main())



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import os
import subprocess
from typing import TextIO


def version_components_from_git(full: bool = False) -> tuple[int, int, int, str]:
    dir = os.getenv("PROJECT_SOURCE_DIR") or os.path.dirname(__file__)
    while True:
        if os.path.exists(os.path.join(dir, ".git")):
            break
        dir = os.path.dirname(dir)
        if dir == os.path.sep:
            raise GitRepoNotFoundError(os.path.dirname(__file__))
    out = subprocess.check_output(
        ["git", "-C", dir, "log", "-1", "--pretty=format:%ad %h", "--date=short"],
        encoding="utf-8",
    )
    parts = out.split()
    date, local = parts[:2]
    major, minor, micro = [int(_) for _ in date.split("-")]
    if full:
        proc = subprocess.Popen(["git", "-C", dir, "diff", "--quiet"])
        proc.wait()
        if proc.returncode:
            local += "-dirty"
    return major - 2000, minor, micro, local


class GitRepoNotFoundError(Exception):
    pass


def write_version_file(file: TextIO, major: int, minor: int, micro: int, local: str) -> None:
    file.write(
        f"""# Version file automatically generated by hpc_connect

__static_version_tuple__: tuple[int, int, int, str] = ({major}, {minor}, {micro}, {local!r})
__static_version__: str = "{major}.{minor}.{micro}+{local}"

def __getattr__(name):

    from hpc_connect.util.dynamic_version import version_components_from_git
    from hpc_connect.util.dynamic_version import GitRepoNotFoundError

    if name not in ("version", "__version__", "version_info", "__version_info__"):
        raise AttributeError(name)
    try:
        major, minor, micro, local = version_components_from_git(full=True)
        if name in ("version", "__version__"):
            return f"{{major}}.{{minor}}.{{micro}}+{{local}}"
        else:
            return (major, minor, micro, local)
    except GitRepoNotFoundError:
        if name in ("version", "__version__"):
            return __static_version__
        else:
            return __static_version_tuple__
"""
    )


def __generate_dynamic_version__():
    major, minor, micro, local = version_components_from_git()
    with open(os.path.join(os.path.dirname(__file__), "../version.py"), "w") as fh:
        write_version_file(fh, major, minor, micro, local)
    if "INCLUDE_LOCAL_VERSION" in os.environ:
        return f"{major}.{minor}.{micro}+{local}"
    return f"{major}.{minor}.{micro}"



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import os

from . import config
from .launch import factory as get_launcher
from .submit import HPCProcess
from .submit import HPCSubmissionFailedError
from .submit import HPCSubmissionManager
from .submit import factory as get_submission_manager


def _initial_logging_setup(*, _ini_setup=[False]):
    if _ini_setup[0]:
        return
    if levelname := os.getenv("HPC_CONNECT_LOG_LEVEL"):
        config.set_logging_level(levelname)
    else:
        config.set_logging_level("INFO")
    if os.getenv("HPC_CONNECT_DEBUG", "no").lower() in ("yes", "true", "1", "on"):
        config.set_logging_level("DEBUG")
    _ini_setup[0] = True


_initial_logging_setup()


# --- backward compatible layer
HPCBackend = HPCSubmissionManager


def get_backend(arg: str) -> HPCSubmissionManager:
    cfg = config.Config()
    cfg.set("submit:backend", arg, scope="internal")
    return get_submission_manager(cfg)



import io
import tokenize
from typing import Generator


def get_tokens(path) -> Generator[tokenize.TokenInfo, None, None]:
    return tokenize.tokenize(io.BytesIO(path.encode("utf-8")).readline)


def strip_quotes(arg: str) -> str:
    s_quote, d_quote = "'''", '"""'
    tokens = get_tokens(arg)
    token = next(tokens)
    while token.type in (tokenize.ENCODING,):
        token = next(tokens)
    s = token.string
    if token.type == tokenize.STRING:
        if s.startswith((s_quote, d_quote)):
            return s[3:-3]
        return s[1:-1]
    return arg



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import datetime
import importlib.resources
import json
import logging
import os
import shutil
import subprocess
from typing import Any

from ..config import Config
from ..hookspec import hookimpl
from .base import HPCProcess
from .base import HPCSubmissionFailedError
from .base import HPCSubmissionManager

logger = logging.getLogger(__name__)


class PBSProcess(HPCProcess):
    def __init__(self, script: str) -> None:
        self._rc: int | None = None
        self._jobid = self.submit(script)
        logger.debug(f"Submitted batch with jobid={self.jobid}")

    def submit(self, script: str) -> str:
        qsub = shutil.which("qsub")
        if qsub is None:
            raise RuntimeError("qsub not found on PATH")
        args = [qsub, script]
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        out, _ = p.communicate()
        result = str(out.decode("utf-8")).strip()
        dirname, basename = os.path.split(script)
        with open(os.path.join(dirname, "qsub.meta.json"), "w") as fh:
            date = datetime.datetime.now().strftime("%c")
            meta = {"args": " ".join(args), "date": date, "stdout/stderr": result}
            json.dump({"meta": meta}, fh, indent=2)
        parts = result.split()
        if len(parts) == 1 and parts[0]:
            return parts[0]
        logger.error("Failed to find jobid!")
        logger.error(f"    The following output was received from {qsub}:")
        for line in result.split("\n"):
            logger.error(f"    {line}")
        raise HPCSubmissionFailedError

    @property
    def jobid(self) -> str:
        return self._jobid

    @property
    def returncode(self) -> int | None:
        return self._rc

    @returncode.setter
    def returncode(self, arg: int) -> None:
        self._rc = arg

    def poll(self) -> int | None:
        qstat = shutil.which("qstat")
        if qstat is None:
            raise RuntimeError("qstat not found on PATH")
        out = subprocess.check_output([qstat], encoding="utf-8")
        lines = [line.strip() for line in out.splitlines() if line.split()]
        for line in lines:
            # Output of qstat is something like:
            # Job id            Name             User              Time Use S Queue
            # ----------------  ---------------- ----------------  -------- - -----
            # 9932285.string-*  spam.sh          username                 0 W serial
            parts = line.split()
            if len(parts) >= 6:
                jid, state = parts[0], parts[4]
                if jid == self.jobid:
                    # Job is still running
                    return None
                elif jid[-1] == "*" and self.jobid.startswith(jid[:-1]):
                    # the output from qstat may return a truncated job id,
                    # so match the beginning of the incoming 'jobids' strings
                    return None
        # Job not found in qstat, assume it completed
        self.returncode = 0
        return self.returncode

    def cancel(self) -> None:
        logger.warning(f"cancelling pbs job {self.jobid}")
        qdel = shutil.which("qdel")
        if qdel is None:
            raise RuntimeError("qdel not found on PATH")
        self.returncode = 1


class PBSSubmissionManager(HPCSubmissionManager):
    """Setup and submit jobs to the PBS scheduler"""

    name = "pbs"

    @staticmethod
    def matches(name: str | None) -> bool:
        return name is not None and name.lower() in ("pbs", "qsub")

    def __init__(self, config: Config | None = None) -> None:
        super().__init__(config=config)
        qsub = shutil.which("qsub")
        if qsub is None:
            raise ValueError("qsub not found on PATH")
        qstat = shutil.which("qstat")
        if qstat is None:
            raise ValueError("qstat not found on PATH")
        qdel = shutil.which("qdel")
        if qdel is None:
            raise ValueError("qdel not found on PATH")

    @property
    def submission_template(self) -> str:
        if "HPCC_PBS_SUBMIT_TEMPLATE" in os.environ:
            return os.environ["HPCC_PBS_SUBMIT_TEMPLATE"]
        return str(importlib.resources.files("hpc_connect").joinpath("templates/pbs.sh.in"))

    def prepare_command_line(self, args: list[str]) -> list[str]:
        qsub = shutil.which("qsub")
        if qsub is None:
            raise ValueError("qsub not found on PATH")
        return [qsub, *self.default_options, *args]

    def submit(
        self,
        name: str,
        args: list[str],
        scriptname: str | None = None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
        **kwargs: Any,
    ) -> PBSProcess:
        cpus = cpus or kwargs.get("tasks")  # backward compatible
        script = self.write_submission_script(
            name,
            args,
            scriptname,
            qtime=qtime,
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        assert script is not None
        return PBSProcess(script)


@hookimpl
def hpc_connect_submission_manager(config) -> HPCSubmissionManager | None:
    if PBSSubmissionManager.matches(config.get("submit:backend")):
        return PBSSubmissionManager(config=config)
    return None



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import os
import re
import stat
from typing import Any
from typing import Callable

from .proc import cpu_count
from .tengine import make_template_env
from .time import hhmmss
from .time import time_in_seconds


def set_executable(path: str) -> None:
    """Set executable bits on ``path``"""
    mode = os.stat(path).st_mode
    if mode & stat.S_IRUSR:
        mode |= stat.S_IXUSR
    if mode & stat.S_IRGRP:
        mode |= stat.S_IXGRP
    if mode & stat.S_IROTH:
        mode |= stat.S_IXOTH
    os.chmod(path, mode)


def partition(arg: list[Any], predicate: Callable) -> tuple[list[Any], list[Any]]:
    a: list[Any] = []
    b: list[Any] = []
    for item in arg:
        if predicate(item):
            a.append(item)
        else:
            b.append(item)
    return a, b


def sanitize_path(path: str) -> str:
    """Remove illegal file characters from ``path``"""
    dirname, basename = os.path.split(path)
    basename = re.sub(r"[^\w_. -]", "_", basename).strip("_")
    return os.path.join(dirname, basename)


def safe_loads(arg):
    import json

    try:
        return json.loads(arg)
    except json.decoder.JSONDecodeError:
        return arg



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import os
import re
import shutil
import subprocess
import sys


def cpu_count(default: int = 4) -> int:
    """Determine the number of processors on the current machine.
    Returns the 'default' if the probes fail.
    """
    if sys.platform == "darwin":
        if cpu_count := read_sysctl():
            return cpu_count
        elif cpu_count := read_lscpu():
            return cpu_count
    else:
        if cpu_count := read_lscpu():
            return cpu_count
        elif cpu_count := read_cpuinfo():
            return cpu_count
    return default


def read_lscpu() -> int | None:
    """"""
    if lscpu := shutil.which("lscpu"):
        try:
            args = [lscpu]
            output = subprocess.check_output(args, encoding="utf-8")
        except subprocess.CalledProcessError:
            return None
        else:
            sockets: int | None = None
            cores_per_socket: int | None = None
            for line in output.split("\n"):
                if line.startswith("Core(s) per socket:"):
                    cores_per_socket = int(line.split(":")[1])
                elif line.startswith("Socket(s):"):
                    sockets = int(line.split(":")[1])
            if cores_per_socket is not None and sockets is not None:
                cpu_count = cores_per_socket * sockets
                return None if cpu_count < 1 else cpu_count
    return None


def read_cpuinfo() -> int | None:
    """
    count the number of lines of this pattern:

        processor       : <integer>
    """
    file = "/proc/cpuinfo"
    if os.path.exists(file):
        proc = re.compile(r"processor\s*:")
        sibs = re.compile(r"siblings\s*:")
        cores = re.compile(r"cpu cores\s*:")
        with open(file, "rt") as fp:
            num_sibs: int = 0
            num_cores: int = 0
            cnt: int = 0
            for line in fp:
                if proc.match(line) is not None:
                    cnt += 1
                elif sibs.match(line) is not None:
                    num_sibs = int(line.split(":")[1])
                elif cores.match(line) is not None:
                    num_cores = int(line.split(":")[1])
            if cnt > 0:
                if num_sibs and num_cores and num_sibs > num_cores:
                    # eg, if num siblings is twice num cores, then physical
                    # cores is half the total processor count
                    fact = int(num_sibs // num_cores)
                    if fact > 0:
                        return cnt // fact
                return cnt
    return None


def read_sysctl():
    if sysctl := shutil.which("sysctl"):
        try:
            args = [sysctl, "-n", "hw.physicalcpu"]
            output = subprocess.check_output(args, encoding="utf-8")
        except subprocess.CalledProcessError:
            return None
        else:
            return int(output)
    return None



# Version file automatically generated by hpc_connect

__static_version_tuple__: tuple[int, int, int, str] = (25, 8, 12, 'f9c4f4f')
__static_version__: str = "25.8.12+f9c4f4f"

def __getattr__(name):

    from hpc_connect.util.dynamic_version import version_components_from_git
    from hpc_connect.util.dynamic_version import GitRepoNotFoundError

    if name not in ("version", "__version__", "version_info", "__version_info__"):
        raise AttributeError(name)
    try:
        major, minor, micro, local = version_components_from_git(full=True)
        if name in ("version", "__version__"):
            return f"{major}.{minor}.{micro}+{local}"
        else:
            return (major, minor, micro, local)
    except GitRepoNotFoundError:
        if name in ("version", "__version__"):
            return __static_version__
        else:
            return __static_version_tuple__



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import argparse
import logging
import math
import os
import shlex
import shutil
import sys
from collections.abc import ValuesView
from functools import cached_property
from typing import IO
from typing import Any

import pluggy
import yaml

from .pluginmanager import HPCConnectPluginManager
from .third_party.schema import Optional
from .third_party.schema import Or
from .third_party.schema import Schema
from .third_party.schema import Use
from .util import collections
from .util import cpu_count
from .util import safe_loads
from .util.string import strip_quotes

logger = logging.getLogger("hpc_connect")


def flag_splitter(arg: list[str] | str) -> list[str]:
    if isinstance(arg, str):
        return shlex.split(arg)
    elif not isinstance(arg, list) and not all(isinstance(str, _) for _ in arg):
        raise ValueError("expected list[str]")
    return arg


def dict_str_str(arg: Any) -> bool:
    f = isinstance
    return f(arg, dict) and all([f(_, str) for k, v in arg.items() for _ in (k, v)])


class choose_from:
    def __init__(self, *choices: str | None):
        self.choices = set(choices)

    def __call__(self, arg: str | None) -> str | None:
        if arg not in self.choices:
            raise ValueError(f"Invalid choice {arg!r}, choose from {self.choices!r}")
        return arg


def which(arg: str) -> str:
    if path := shutil.which(arg):
        return path
    logger.debug(f"{arg} not found on PATH")
    return arg


resource_spec = {
    "type": "node",
    "count": int,
    "resources": [
        {
            "type": str,
            "count": int,
            Optional("resources"): [
                {
                    "type": str,
                    "count": int,
                },
            ],
        },
    ],
}


launch_spec = {
    Optional("numproc_flag"): str,
    Optional("default_options"): Use(flag_splitter),
    Optional("local_options"): Use(flag_splitter),
    Optional("pre_options"): Use(flag_splitter),
    Optional("mappings"): dict_str_str,
}

schema = Schema(
    {
        "hpc_connect": {
            Optional("config"): {
                Optional("debug"): bool,
            },
            Optional("submit"): {
                Optional("backend"): Use(
                    choose_from(None, "shell", "slurm", "sbatch", "pbs", "qsub", "flux")
                ),
                Optional("default_options"): Use(flag_splitter),
                Optional(str): {
                    Optional("default_options"): Use(flag_splitter),
                },
            },
            Optional("machine"): {
                Optional("resources"): Or([resource_spec], None),
            },
            Optional("launch"): {
                Optional("exec"): Use(which),
                **launch_spec,
                Optional(str): launch_spec,
            },
        }
    },
    ignore_extra_keys=True,
    description="HPC connect configuration schema",
)


class ConfigScope:
    def __init__(self, name: str, file: str | None, data: dict[str, Any]) -> None:
        self.name = name
        self.file = file
        self.data = schema.validate({"hpc_connect": data})["hpc_connect"]

    def __repr__(self):
        file = self.file or "<none>"
        return f"ConfigScope({self.name}: {file})"

    def __eq__(self, other):
        if not isinstance(other, ConfigScope):
            return False
        return self.name == other.name and self.file == other.file and other.data == self.data

    def __iter__(self):
        return iter(self.data)

    def get_section(self, section: str) -> Any:
        return self.data.get(section)

    def dump(self) -> None:
        if self.file is None:
            return
        with open(self.file, "w") as fh:
            yaml.dump({"hpc_connect": self.data}, fh, default_flow_style=False)


config_defaults = {
    "config": {
        "debug": False,
    },
    "machine": {
        "resources": None,
    },
    "submit": {
        "backend": None,
        "default_options": [],
    },
    "launch": {
        "exec": "mpiexec",
        "numproc_flag": "-n",
        "default_options": [],
        "local_options": [],
        "pre_options": [],
        "mappings": {},
    },
}


class Config:
    def __init__(self) -> None:
        self.pluginmanager: pluggy.PluginManager = HPCConnectPluginManager()
        self.scopes: dict[str, ConfigScope] = {
            "defaults": ConfigScope("defaults", None, config_defaults)
        }
        for scope in ("site", "global", "local"):
            config_scope = read_config_scope(scope)
            self.push_scope(config_scope)
        if cscope := read_env_config():
            self.push_scope(cscope)
        if self.get("config:debug"):
            set_logging_level("debug")

    def read_only_scope(self, scope: str) -> bool:
        return scope in ("defaults", "environment", "command_line")

    def push_scope(self, scope: ConfigScope) -> None:
        self.scopes[scope.name] = scope

    def pop_scope(self, scope: ConfigScope) -> ConfigScope | None:
        return self.scopes.pop(scope.name, None)

    def get_config(self, section: str, scope: str | None = None) -> Any:
        scopes: ValuesView[ConfigScope] | list[ConfigScope]
        if scope is None:
            scopes = self.scopes.values()
        else:
            scopes = [self.validate_scope(scope)]
        merged_section: dict[str, Any] = {}
        for config_scope in scopes:
            data = config_scope.get_section(section)
            if not data or not isinstance(data, dict):
                continue
            merged_section = collections.merge(merged_section, {section: data})
        if section not in merged_section:
            return {}
        return merged_section[section]

    def get(self, path: str, default: Any = None, scope: str | None = None) -> Any:
        parts = process_config_path(path)
        section = parts.pop(0)
        value = self.get_config(section, scope=scope)
        while parts:
            key = parts.pop(0)
            # cannot use value.get(key, default) in case there is another part
            # and default is not a dict
            if key not in value:
                return default
            value = value[key]
        return value

    def set(self, path: str, value: Any, scope: str | None = None) -> None:
        parts = process_config_path(path)
        section = parts.pop(0)
        section_data = self.get_config(section, scope=scope)
        data = section_data
        while len(parts) > 1:
            key = parts.pop(0)
            new = data.get(key, {})
            if isinstance(new, dict):
                new = dict(new)
                # reattach to parent object
                data[key] = new
            data = new
        # update new value
        data[parts[0]] = value
        self.update_config(section, section_data, scope=scope)

    def add(self, fullpath: str, scope: str | None = None) -> None:
        path: str = ""
        existing: Any = None
        components = process_config_path(fullpath)
        has_existing_value = True
        for idx, name in enumerate(components[:-1]):
            path = name if not path else f"{path}:{name}"
            existing = self.get(path, scope=scope)
            if existing is None:
                has_existing_value = False
                # construct value from this point down
                value = safe_loads(components[-1])
                for component in reversed(components[idx + 1 : -1]):
                    value = {component: value}
                break

        if has_existing_value:
            path = ":".join(components[:-1])
            value = safe_loads(strip_quotes(components[-1]))
            existing = self.get(path, scope=scope)

        if isinstance(existing, list) and not isinstance(value, list):
            # append values to lists
            value = [value]

        new = collections.merge(existing, value)
        self.set(path, new, scope=scope)

    def highest_precedence_scope(self) -> ConfigScope:
        """Non-internal scope with highest precedence."""
        file_scopes = [scope for scope in self.scopes.values() if scope.file is not None]
        return next(reversed(file_scopes))

    def validate_scope(self, scope: str | None) -> ConfigScope:
        if scope is None:
            return self.highest_precedence_scope()
        elif scope in self.scopes:
            return self.scopes[scope]
        elif scope == "internal":
            self.scopes["internal"] = ConfigScope("internal", None, {})
            return self.scopes["internal"]
        else:
            raise ValueError(f"Invalid scope {scope!r}")

    def update_config(self, section: str, update_data: dict[str, Any], scope: str | None = None):
        """Update the configuration file for a particular scope.

        Args:
            section (str): section of the configuration to be updated
            update_data (dict): data to be used for the update
            scope (str): scope to be updated
        """
        if scope is None:
            config_scope = self.highest_precedence_scope()
        else:
            config_scope = self.scopes[scope]
        # read only the requested section's data.
        config_scope.data[section] = dict(update_data)
        config_scope.dump()

    def set_main_options(self, args: argparse.Namespace) -> None:
        """Set main configuration options based on command-line arguments.

        Updates the configuration attributes based on the provided argparse Namespace containing
        command-line arguments.

        Args:
            args: An argparse.Namespace object containing command-line arguments.
        """
        if args.config_mods:
            data: dict[str, Any] = {}
            for fullpath in args.config_mods:
                current = data
                components = process_config_path(fullpath)
                for component in components[:-2]:
                    current = current.setdefault(component, {})
                current[components[-2]] = safe_loads(components[-1])
            scope = ConfigScope("command_line", None, data)
            self.push_scope(scope)
            if self.get("config:debug", scope="command_line"):
                set_logging_level("debug")

    @property
    def resource_specs(self) -> list[dict]:
        from .submit import factory

        if resource_specs := self.get("machine:resources"):
            return resource_specs
        if self.get("submit:backend"):
            # backend may set resources
            factory(config=self)
        if resource_specs := self.get("machine:resources"):
            return resource_specs
        resource_specs = default_resource_spec()
        self.set("machine:resources", resource_specs, scope="defaults")
        return resource_specs

    def count_per_rspec(self, rspec: dict[str, Any], type: str) -> int | None:
        for child in rspec["resources"]:
            if child["type"] == type:
                return child["count"]
        return None

    def count_per_node(self, type: str) -> int:
        for rspec in self.resource_specs:
            if rspec["type"] == "node":
                count = self.count_per_rspec(rspec, type)
                if count is not None:
                    return count
        raise ValueError(f"Unable to determine count_per_node for {type!r}")

    def count_per_socket(self, type: str) -> int:
        for rspec1 in self.resource_specs:
            if rspec1["type"] == "node":
                for rspec2 in rspec1["resources"]:
                    if rspec2["type"] == "socket":
                        count = self.count_per_rspec(rspec2, type)
                        if count is not None:
                            return count
        raise ValueError(f"Unable to determine count_per_socket for {type!r}")

    @cached_property
    def node_count(self) -> int:
        for resource in self.resource_specs:
            if resource["type"] == "node":
                return resource["count"]
        raise ValueError("Unable to determine node count")

    @cached_property
    def sockets_per_node(self) -> int:
        try:
            count = self.count_per_node("socket")
            return count or 1
        except ValueError:
            return 1

    @cached_property
    def cpus_per_socket(self) -> int:
        try:
            return self.count_per_socket("cpu")
        except ValueError:
            return cpu_count()

    @cached_property
    def gpus_per_socket(self) -> int:
        try:
            return self.count_per_socket("gpu")
        except ValueError:
            return 0

    @cached_property
    def cpus_per_node(self) -> int:
        return self.sockets_per_node * self.cpus_per_socket

    @cached_property
    def gpus_per_node(self) -> int:
        if gpus_per_resource_type := self.gpus_per_socket:
            return self.sockets_per_node * gpus_per_resource_type
        try:
            return self.count_per_node("gpu")
        except ValueError:
            return 0

    @cached_property
    def cpu_count(self) -> int:
        return self.node_count * self.cpus_per_node

    @cached_property
    def gpu_count(self) -> int:
        return self.node_count * self.gpus_per_node

    def nodes_required(self, max_cpus: int | None = None, max_gpus: int | None = None) -> int:
        """Nodes required to run ``tasks`` tasks.  A task can be thought of as a single MPI
        rank"""
        nodes = max(1, int(math.ceil((max_cpus or 1) / self.cpus_per_node)))
        if self.gpus_per_node:
            nodes = max(nodes, int(math.ceil((max_gpus or 0) / self.gpus_per_node)))
        return nodes

    def compute_required_resources(
        self, *, ranks: int | None = None, ranks_per_socket: int | None = None
    ) -> dict[str, int]:
        """Return basic information about how to allocate resources on this machine for a job
        requiring `ranks` ranks.

        Parameters
        ----------
        ranks : int
            The number of ranks to use for a job
        ranks_per_socket : int
            Number of ranks per socket, for performance use

        Returns
        -------
        SimpleNamespace

        """
        if ranks is None and ranks_per_socket is not None:
            # Raise an error since there is no reliable way of finding the number of
            # available nodes
            raise ValueError("ranks_per_socket requires ranks also be defined")

        reqd_resources: dict[str, int] = {
            "np": 0,
            "ranks": 0,
            "ranks_per_socket": 0,
            "nodes": 0,
            "sockets": 0,
        }

        if not ranks and not ranks_per_socket:
            return reqd_resources

        nodes: int
        if ranks is None and ranks_per_socket is None:
            ranks = ranks_per_socket = 1
            nodes = 1
        elif ranks is not None and ranks_per_socket is None:
            ranks_per_socket = min(ranks, self.cpus_per_socket)
            nodes = int(math.ceil(ranks / self.cpus_per_socket / self.sockets_per_node))
        else:
            assert ranks is not None
            assert ranks_per_socket is not None
            nodes = int(math.ceil(ranks / ranks_per_socket / self.sockets_per_node))
        sockets = int(math.ceil(ranks / ranks_per_socket))
        reqd_resources["np"] = ranks
        reqd_resources["ranks"] = ranks
        reqd_resources["ranks_per_socket"] = ranks_per_socket
        reqd_resources["nodes"] = nodes
        reqd_resources["sockets"] = sockets
        return reqd_resources

    def dump(self, stream: IO[Any], scope: str | None = None, **kwargs: Any) -> None:
        from .submit import factory

        # initialize the resource spec
        if self.get("machine:resources") is None:
            if self.get("submit:backend"):
                factory(self)
            if not self.get("machine:resources"):
                self.set("machine:resources", default_resource_spec(), scope="defaults")
        data: dict[str, Any] = {}
        for section in self.scopes["defaults"]:
            section_data = self.get_config(section, scope=scope)
            if not section_data and scope is not None:
                continue
            data[section] = section_data
        yaml.dump({"hpc_connect": data}, stream, **kwargs)


def read_config_scope(scope: str) -> ConfigScope:
    data: dict[str, Any] = {}
    if file := get_scope_filename(scope):
        if fd := read_config_file(file):
            if "hpc_connect" not in fd:
                raise KeyError("Missing key 'hpc_connect'")
            data.update(fd["hpc_connect"])
    return ConfigScope(scope, file, data)


def get_scope_filename(scope: str) -> str | None:
    if scope == "site":
        if var := os.getenv("HPC_CONNECT_SITE_CONFIG"):
            return var
        return os.path.join(sys.prefix, "etc/hpc_connect/config.yaml")
    elif scope == "global":
        if var := os.getenv("HPC_CONNECT_GLOBAL_CONFIG"):
            return var
        elif var := os.getenv("XDG_CONFIG_HOME"):
            file = os.path.join(var, "hpc_connect/config.yaml")
            if os.path.exists(file):
                return file
        return os.path.expanduser("~/.config/hpc_connect.yaml")
    elif scope == "local":
        return os.path.abspath("./hpc_connect.yaml")
    raise ValueError(f"Could not determine filename for scope {scope!r}")


def read_env_config() -> ConfigScope | None:
    def load_mappings(arg: str) -> dict[str, str]:
        mappings: dict[str, str] = {}
        for kv in arg.split(","):
            k, v = [_.strip() for _ in kv.split(":") if _.split()]
            mappings[k] = v
        return mappings

    data: dict[str, Any] = {}
    for var in os.environ:
        if not var.startswith("HPCC_"):
            continue
        try:
            section, *parts = var[5:].lower().split("_")
            key = "_".join(parts)
        except ValueError:
            continue
        if section not in config_defaults:
            continue
        value: Any
        if key == "mappings":
            value = load_mappings(os.environ[var])
        else:
            value = safe_loads(os.environ[var])
        data.setdefault(section, {}).update({key: value})
    if not data:
        return None
    return ConfigScope("environment", None, data)


def read_config_file(file: str) -> dict[str, Any] | None:
    """Load configuration settings from ``file``"""
    if not os.path.exists(file):
        return None
    with open(file) as fh:
        return yaml.safe_load(fh)


def process_config_path(path: str) -> list[str]:
    result: list[str] = []
    if path.startswith(":"):
        raise ValueError(f"Illegal leading ':' in path {path}")
    while path:
        front, _, path = path.partition(":")
        result.append(front)
        if path.startswith(("{", "[")):
            result.append(path)
            return result
    return result


ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter("==> %(message)s"))
logger.addHandler(ch)


loglevelmap: dict[str, int] = {
    "CRITICAL": logging.CRITICAL,
    "FATAL": logging.FATAL,
    "ERROR": logging.ERROR,
    "WARN": logging.WARNING,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}


def set_debug(arg: bool) -> None:
    if arg:
        set_logging_level("DEBUG")


def set_logging_level(levelname: str) -> None:
    level = loglevelmap[levelname.upper()]
    for h in logger.handlers:
        h.setLevel(level)
    logger.setLevel(level)


def default_resource_spec() -> list[dict]:
    resource_spec: list[dict] = [
        {
            "type": "node",
            "count": 1,
            "resources": [
                {
                    "type": "socket",
                    "count": 1,
                    "resources": [
                        {
                            "type": "cpu",
                            "count": cpu_count(),
                        },
                    ],
                },
            ],
        }
    ]
    return resource_spec



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import getpass
import importlib
import os
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from typing import Any
from typing import Protocol
from typing import TextIO

from ..config import Config
from ..util import make_template_env
from ..util import sanitize_path
from ..util import set_executable
from ..util import time_in_seconds


class HPCProcess(Protocol):
    @property
    def jobid(self) -> Any:
        return "<none>"

    @property
    def returncode(self) -> int | None:
        raise NotImplementedError

    @returncode.setter
    def returncode(self, arg: int) -> None:
        raise NotImplementedError

    def poll(self) -> int | None:
        raise NotImplementedError

    def cancel(self) -> None:
        raise NotImplementedError


class HPCSubmissionManager(ABC):
    name = "<backend>"

    def __init__(self, config: Config | None = None) -> None:
        self.config = config or Config()

    @staticmethod
    @abstractmethod
    def matches(name: str) -> bool: ...

    @property
    def supports_subscheduling(self) -> bool:
        return False

    def get_from_config(self, key: str, default: Any = None) -> Any:
        if value := self.config.get(f"submit:{self.name}:{key}"):
            return value
        else:
            value = self.config.get(f"submit:{key}")
            return default if value is None else value

    @property
    def default_options(self) -> list[str]:
        return self.get_from_config("default_options") or []

    @property
    def polling_frequency(self) -> float:
        s = os.getenv("HPCC_POLLING_FREQUENCY") or 30.0  # 30s.
        return time_in_seconds(s)

    def prepare_command_line(self, args: list[str]) -> list[str]:
        raise NotImplementedError

    @abstractmethod
    def submit(
        self,
        name: str,
        args: list[str],
        scriptname: str | None = None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
        **kwargs: Any,
    ) -> HPCProcess: ...

    def submitn(
        self,
        name: list[str],
        args: list[list[str]],
        scriptname: list[str] | None = None,
        qtime: list[float] | None = None,
        submit_flags: list[list[str]] | None = None,
        variables: list[dict[str, str | None]] | None = None,
        output: list[str] | None = None,
        error: list[str] | None = None,
        nodes: list[int] | None = None,
        cpus: list[int] | None = None,
        gpus: list[int] | None = None,
        **kwargs: Any,
    ) -> HPCProcess:
        raise NotImplementedError(f"{self.name} backend has not implemented submitn")

    @property
    def submission_template(self) -> str:
        raise NotImplementedError

    def write_submission_script(
        self,
        name: str,
        args: list[str],
        scriptname: str | TextIO | None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        #
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
    ) -> str | None:
        template_dirs = {str(importlib.resources.files("hpc_connect").joinpath("templates"))}
        data = self.format_submission_data(
            name,
            args,
            qtime=qtime,
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        template = self.submission_template
        if not os.path.exists(template):
            raise FileNotFoundError(template)
        d, f = os.path.split(os.path.abspath(template))
        template_dirs.add(d)
        env = make_template_env(*template_dirs)
        t = env.get_template(f)
        if hasattr(scriptname, "write"):
            scriptname.write(t.render(data))  # type: ignore
            return None
        scriptname = scriptname or unique_scriptname()
        file = sanitize_path(scriptname)
        os.makedirs(os.path.dirname(file), exist_ok=True)
        with open(file, "w") as fh:
            fh.write(t.render(data))
        set_executable(file)
        return file

    def format_submission_data(
        self,
        name: str,
        args: list[str],
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        #
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
    ) -> dict[str, Any]:
        if nodes is None and cpus is None:
            raise ValueError("must specify at least one of nodes and cpus")
        output = output or "stdout.txt"
        data: dict[str, Any] = {
            "name": name,
            "time": qtime or 1.0,
            "args": submit_flags or [],
            "nodes": nodes,
            "cpus": cpus,
            "gpus": gpus,
            "cpus_per_node": self.config.cpus_per_node,
            "gpus_per_node": self.config.gpus_per_node,
            "user": getpass.getuser(),
            "date": datetime.now().strftime("%c"),
            "variables": variables or {},
            "commands": args,
            "output": output,
            "error": error,
        }
        return data


def unique_scriptname() -> str:
    template = "Submit-%d.sh"
    i = 1
    while True:
        script = template % i
        if not os.path.exists(script):
            return script
        i += 1


class HPCSubmissionFailedError(Exception):
    pass



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import io
import os
from contextlib import contextmanager

import hpc_connect


@contextmanager
def tmp_environ():
    save_env = os.environ.copy()
    try:
        os.environ["HPC_CONNECT_CPUS_PER_NODE"] = "10"
        os.environ["HPC_CONNECT_GPUS_PER_NODE"] = "0"
        os.environ["HPC_CONNECT_NODE_COUNT"] = "1"
        yield
    finally:
        os.environ.clear()
        os.environ.update(save_env)


def test_basic():
    backend = hpc_connect.submit.pbs.PBSSubmissionManager()
    with io.StringIO() as fh:
        backend.write_submission_script(
            "my-job",
            ["ls"],
            fh,
            cpus=1,
            nodes=1,
            output="my-out.txt",
            error="my-err.txt",
            qtime=1.0,
            variables={"MY_VAR": "SPAM"},
        )
        text = fh.getvalue()
    assert "#!/bin/sh" in text
    assert "#PBS -N my-job" in text
    assert f"#PBS -l nodes=1:ppn={backend.config.cpus_per_node}" in text
    assert "#PBS -l walltime=00:00:01" in text
    assert "#PBS --job-name=my-job" in text
    assert "#PBS -o my-out.txt" in text
    assert "#PBS -e my-err.txt" in text
    assert "export MY_VAR=SPAM" in text
    assert "printenv || true" in text
    assert "ls" in text



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import argparse
import os
import shlex

from ..config import Config

description = "Wrapper to submit tool, such as sbatch"
add_help = False


def setup_parser(parser: argparse.ArgumentParser) -> None:
    from .. import submit

    parser.usage = "%(prog)s [-h] [--help] ..."
    parser.add_argument(
        "-h",
        default=False,
        action="help",
        help="Show this message and exit.",
    )
    parser.add_argument(
        "--dryrun",
        action="store_true",
        default=False,
        help="Print command line to shell but do not execute",
    )
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.epilog = submit.__doc__


def execute(config: Config, args: argparse.Namespace) -> None:
    import hpc_connect

    argv = list(args.extra_args)
    manager = hpc_connect.get_submission_manager(config)
    cmd = manager.prepare_command_line(argv)
    if args.dryrun:
        print(shlex.join(cmd))
        return
    os.execvp(cmd[0], cmd)



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import importlib
import logging
import os
import shutil
import subprocess
import weakref
from typing import Any
from typing import TextIO

import psutil

from ..config import Config
from ..hookspec import hookimpl
from ..util import time_in_seconds
from .base import HPCProcess
from .base import HPCSubmissionManager

logger = logging.getLogger(__name__)


def streamify(arg: str | None) -> TextIO | None:
    if arg is None:
        return None
    os.makedirs(os.path.dirname(arg), exist_ok=True)
    return open(arg, mode="w")


class ShellProcess(HPCProcess):
    def __init__(
        self,
        script: str,
        output: str | None = None,
        error: str | None = None,
    ) -> None:
        sh = shutil.which("sh")
        if sh is None:
            raise RuntimeError("sh not found on PATH")
        stdout = streamify(output)
        if stdout is not None:
            weakref.finalize(stdout, stdout.close)
        stderr: TextIO | int | None
        if error is None:
            stderr = None
        elif error == output:
            stderr = subprocess.STDOUT
        else:
            stderr = streamify(error)
        if hasattr(stderr, "write"):
            weakref.finalize(stderr, stderr.close)  # type: ignore
        self.proc = subprocess.Popen([sh, script], stdout=stdout, stderr=stderr)

    @property
    def returncode(self) -> int | None:
        return self.proc.returncode

    @returncode.setter
    def returncode(self, arg: int) -> None:
        raise NotImplementedError

    def poll(self) -> int | None:
        return self.proc.poll()

    def cancel(self) -> None:
        """Kill a process tree (including grandchildren)"""
        logger.warning(f"cancelling shell batch with pid {self.proc.pid}")
        try:
            parent = psutil.Process(self.proc.pid)
        except psutil.NoSuchProcess:
            return
        children = parent.children(recursive=True)
        children.append(parent)
        for p in children:
            try:
                p.terminate()
            except Exception:
                pass
        _, alive = psutil.wait_procs(children)
        for p in alive:
            try:
                p.kill()
            except Exception:
                pass


class ShellSubmissionManager(HPCSubmissionManager):
    name = "shell"

    def __init__(self, config: Config | None = None):
        super().__init__(config=config)
        sh = shutil.which("sh")
        if sh is None:
            raise ValueError("sh not found on PATH")

    @staticmethod
    def matches(name) -> bool:
        return name in ("shell", "none")

    @property
    def polling_frequency(self) -> float:
        s = os.getenv("HPCC_POLLING_FREQUENCY") or 0.5
        return time_in_seconds(s)

    def submit(
        self,
        name: str,
        args: list[str],
        scriptname: str | None = None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
        **kwargs: Any,
    ) -> HPCProcess:
        cpus = cpus or kwargs.get("tasks")  # backward compatible
        script = self.write_submission_script(
            name,
            args,
            scriptname,
            qtime=qtime,
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        assert script is not None
        return ShellProcess(script, output=output, error=error)

    @property
    def submission_template(self) -> str:
        return str(importlib.resources.files("hpc_connect").joinpath("templates/shell.sh.in"))


@hookimpl
def hpc_connect_submission_manager(config) -> HPCSubmissionManager | None:
    if ShellSubmissionManager.matches(config.get("submit:backend")):
        return ShellSubmissionManager(config=config)
    return None



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import argparse
import datetime
import importlib.resources
import json
import logging
import os
import re
import shutil
import subprocess
import time
from typing import Any

from ..config import Config
from ..hookspec import hookimpl
from .base import HPCProcess
from .base import HPCSubmissionFailedError
from .base import HPCSubmissionManager

logger = logging.getLogger(__name__)


class SlurmProcess(HPCProcess):
    def __init__(self, script: str) -> None:
        self._rc: int | None = None
        self.clusters: str | None = None
        self.script = os.path.abspath(script)
        self.script_dir = os.path.dirname(self.script)
        self._jobid = self.submit(script)
        f = os.path.basename(self.script)
        logger.debug(f"Submitted batch script {f} with jobid={self.jobid}")

    @property
    def jobid(self) -> str:
        return self._jobid

    def submit(self, script: str) -> str:
        sbatch = shutil.which("sbatch")
        if sbatch is None:
            raise ValueError("sbatch not found on PATH")
        ns = self.parse_script_args(script)
        if ns.clusters:
            self.clusters = ns.clusters
        args = [sbatch, script]
        proc = subprocess.run(args, check=True, encoding="utf-8", capture_output=True)
        with open(os.path.join(self.script_dir, "submit.meta.json"), "w") as fh:
            date = datetime.datetime.now().strftime("%c")
            meta = {"args": " ".join(args), "date": date, "stdout/stderr": proc.stdout}
            json.dump({"meta": meta}, fh, indent=2)
        if match := re.search("Submitted batch job (.*)$", proc.stdout):
            jobid = match.group(1).strip()
            return jobid
        logger.error(f"Failed to find jobid!\n    The following output was received from {sbatch}:")
        for line in proc.stdout.split("\n"):
            logger.log(logging.ERROR, f"    {line}")
        for line in proc.stderr.split("\n"):
            logger.log(logging.ERROR, f"    {line}")
        raise HPCSubmissionFailedError

    @staticmethod
    def parse_script_args(script: str) -> argparse.Namespace:
        args = []
        with open(script, "r") as file:
            for line in file:
                if match := re.search("^#SBATCH\s+(.*)$", line):
                    args.append(match.group(1).strip())
        p = argparse.ArgumentParser()
        p.add_argument("-M", "--cluster", "--clusters", dest="clusters")
        ns, _ = p.parse_known_args(args)
        return ns

    @property
    def returncode(self) -> int | None:
        return self._rc

    @returncode.setter
    def returncode(self, arg: int) -> None:
        self._rc = arg

    def poll(self) -> int | None:
        sacct = shutil.which("sacct")
        if sacct is None:
            raise RuntimeError("sacct not found on PATH")
        max_tries: int = 20
        acct_data: dict[str, dict[str, Any]] = {}
        for _ in range(max_tries):
            args = [sacct, "--noheader", "-j", self.jobid, "-p", "-b"]
            if self.clusters:
                args.append(f"--clusters={self.clusters}")
            proc = subprocess.run(args, check=True, encoding="utf-8", capture_output=True)
            lines = [line.strip() for line in proc.stdout.splitlines() if line.split()]
            if lines:
                for line in lines:
                    jobid, state, exit_code = [_.strip() for _ in line.split("|") if _.split()]
                    try:
                        returncode, signal = [int(_) for _ in exit_code.split(":")]
                    except ValueError:
                        returncode = int(exit_code)
                        signal = 0
                    acct_data[jobid] = {
                        "state": state.split()[0].rstrip("+"),
                        "returncode": returncode,
                        "signal": signal,
                    }
                break
            time.sleep(0.5)
        else:
            raise RuntimeError(f"{' '.join(args)!r} did not return any accounting data")

        if jobinfo := acct_data.get(self.jobid):
            if jobinfo["state"].upper() in ("PENDING", "RUNNING"):
                return None
            self.returncode = max(jobinfo["returncode"], jobinfo["signal"])
            if jobinfo["signal"]:
                logger.error(f"Job {self.jobid} failed with signal {jobinfo['signal']}")
                f = os.path.join(self.script_dir, f"{self.jobid}.acct.json")
                with open(f, "w") as fh:
                    args = [sacct, "-j", self.jobid, "--json"]
                    subprocess.run(args, stdout=fh, encoding="utf-8")
        else:
            raise RuntimeError(f"Accounting data for job {self.jobid} not returned by sacct")
        return self.returncode

    def cancel(self) -> None:
        logger.warning(f"cancelling slurm job {self.jobid}")
        subprocess.run(["scancel", self.jobid, "--clusters=all"])
        self.returncode = 1


class SlurmSubmissionManager(HPCSubmissionManager):
    """Setup and submit jobs to the slurm scheduler"""

    name = "slurm"

    @staticmethod
    def matches(name: str | None) -> bool:
        return name is not None and name.lower() in ("slurm", "sbatch")

    def __init__(self, config: Config | None = None) -> None:
        super().__init__(config=config)
        sbatch = shutil.which("sbatch")
        if sbatch is None:
            raise ValueError("sbatch not found on PATH")
        sacct = shutil.which("sacct")
        if sacct is None:
            raise ValueError("sacct not found on PATH")
        if self.config.get("machine:resources") is None:
            if sinfo := read_sinfo():
                self.config.set("machine:resources", [sinfo], scope="defaults")
        else:
            logger.warning("Unable to determine system configuration from sinfo, using default")

    @property
    def submission_template(self) -> str:
        if "HPCC_SLURM_SUBMIT_TEMPLATE" in os.environ:
            return os.environ["HPCC_SLURM_SUBMIT_TEMPLATE"]
        return str(importlib.resources.files("hpc_connect").joinpath("templates/slurm.sh.in"))

    def prepare_command_line(self, args: list[str]) -> list[str]:
        sbatch = shutil.which("sbatch")
        if sbatch is None:
            raise ValueError("sbatch not found on PATH")
        return [sbatch, *self.default_options, *args]

    def submit(
        self,
        name: str,
        args: list[str],
        scriptname: str | None = None,
        qtime: float | None = None,
        submit_flags: list[str] | None = None,
        variables: dict[str, str | None] | None = None,
        output: str | None = None,
        error: str | None = None,
        nodes: int | None = None,
        cpus: int | None = None,
        gpus: int | None = None,
        **kwargs: Any,
    ) -> SlurmProcess:
        cpus = cpus or kwargs.get("tasks")  # backward compatible
        script = self.write_submission_script(
            name,
            args,
            scriptname,
            qtime=qtime,
            submit_flags=submit_flags,
            variables=variables,
            output=output,
            error=error,
            nodes=nodes,
            cpus=cpus,
            gpus=gpus,
        )
        assert script is not None
        return SlurmProcess(script)


def read_sinfo() -> dict[str, Any] | None:
    if sinfo := shutil.which("sinfo"):
        opts = [
            "%X",  # Number of sockets per node
            "%Y",  # Number of cores per socket
            "%Z",  # Number of threads per core
            "%c",  # Number of CPUs per node
            "%D",  # Number of nodes
            "%G",  # General resources
        ]
        format = " ".join(opts)
        args = [sinfo, "-o", format]
        try:
            proc = subprocess.run(args, check=True, encoding="utf-8", capture_output=True)
        except subprocess.CalledProcessError:
            return None
        else:
            for line in proc.stdout.split("\n"):
                parts = line.split()
                if not parts:
                    continue
                elif parts and parts[0].startswith("SOCKETS"):
                    continue
                spn, cps, _, cpn, nc, *gres = [safe_loads(_) for _ in parts]
                break
            else:
                raise ValueError(f"Unable to read sinfo output:\n{proc.stdout}")
            info: dict[str, Any] = {
                "type": "node",
                "count": nc,
                "resources": [
                    {
                        "type": "socket",
                        "count": spn,
                        "resources": [
                            {
                                "type": "cpu",
                                "count": cps,
                            },
                        ],
                    }
                ],
            }
            for res in gres:
                if not res:
                    continue
                parts = res.split(":")
                resource: dict[str, Any] = {
                    "type": parts[0],
                    "count": safe_loads(parts[-1]),
                }
                if len(parts) > 2:
                    resource["gres"] = ":".join(parts[1:-1])
                info["resources"].append(resource)
            return info
    return None


def safe_loads(arg: str) -> Any:
    if arg == "(null)":
        return None
    if arg.endswith("+"):
        return safe_loads(arg[:-1])
    try:
        return json.loads(arg)
    except json.JSONDecodeError:
        return arg


@hookimpl
def hpc_connect_submission_manager(config) -> HPCSubmissionManager | None:
    if SlurmSubmissionManager.matches(config.get("submit:backend")):
        return SlurmSubmissionManager(config=config)
    return None



from ..config import Config
from . import flux_hl
from . import pbs
from . import shell
from . import slurm
from .base import HPCProcess
from .base import HPCSubmissionFailedError
from .base import HPCSubmissionManager

plugins = (flux_hl, pbs, shell, slurm)


def factory(config: Config | None = None) -> HPCSubmissionManager:
    config = config or Config()
    submission_manager = config.pluginmanager.hook.hpc_connect_submission_manager(config=config)
    if submission_manager is None:
        backend = config.get("submit:backend")
        raise ValueError(f"No matching submission manager for {backend!r}")
    return submission_manager



import os
from contextlib import contextmanager

import yaml

import hpc_connect


@contextmanager
def working_dir(dirname):
    save_cwd = os.getcwd()
    os.makedirs(dirname, exist_ok=True)
    try:
        os.chdir(dirname)
        yield
    finally:
        os.chdir(save_cwd)


@contextmanager
def envmods(**kwargs):
    try:
        save_env = os.environ.copy()
        for key in os.environ:
            if key.startswith("HPCC_"):
                os.environ.pop(key)
        os.environ.update(kwargs)
        yield
    finally:
        os.environ.clear()
        os.environ.update(save_env)


mock_bin = os.path.join(os.path.dirname(__file__), "mock")


def launch(args, **kwargs):
    launcher = hpc_connect.get_launcher()
    launcher(args, **kwargs)


def test_envar_config(capfd):
    env = {"HPCC_LAUNCH_EXEC": "srun", "HPCC_LAUNCH_NUMPROC_FLAG": "-np"}
    with envmods(**env):
        launch(["-n", "4", "-flag", "file", "executable", "--option"])
        captured = capfd.readouterr()
        out = captured.out.strip()
        assert out == f"{mock_bin}/srun -np 4 -flag file executable --option"
    env = {
        "HPCC_LAUNCH_EXEC": "mpiexec",
        "HPCC_LAUNCH_NUMPROC_FLAG": "-np",
        "HPCC_LAUNCH_LOCAL_OPTIONS": "--map-by ppr:%(np)d:cores",
    }
    with envmods(**env):
        launch(["-n", "4", "-flag", "file", "executable", "--option"])
        captured = capfd.readouterr()
        out = captured.out.strip()
        assert out == f"{mock_bin}/mpiexec --map-by ppr:4:cores -np 4 -flag file executable --option"


def test_file_config(tmpdir, capfd):
    with envmods(HPC_CONNECT_GLOBAL_CONFIG="hpc_connect.yaml"):
        with working_dir(str(tmpdir)):
            with open("hpc_connect.yaml", "w") as fh:
                yaml.dump({"hpc_connect": {"launch": {"exec": "srun", "numproc_flag": "-np"}}}, fh)
            launch(["-n", "4", "-flag", "file", "executable", "--option"])
            captured = capfd.readouterr()
            out = captured.out.strip()
            assert out == f"{mock_bin}/srun -np 4 -flag file executable --option"

            with open("hpc_connect.yaml", "w") as fh:
                yaml.dump(
                    {
                        "hpc_connect": {
                            "launch": {
                                "exec": "mpiexec",
                                "numproc_flag": "-np",
                                "local_options": "--map-by ppr:%(np)d:cores",
                            }
                        }
                    },
                    fh,
                )
            launch(["-n", "4", "-flag", "file", "executable", "--option"])
            captured = capfd.readouterr()
            out = captured.out.strip()
            assert out == f"{mock_bin}/mpiexec --map-by ppr:4:cores -np 4 -flag file executable --option"

            with open("hpc_connect.yaml", "w") as fh:
                yaml.dump(
                    {
                        "hpc_connect": {
                            "launch": {
                                "exec": "mpiexec",
                                "numproc_flag": "-np",
                                "mappings": {"-flag": "-xflag"},
                            }
                        }
                    },
                    fh,
                )

            launch(["-n", "4", "-flag", "file", "executable", "--option"])
            captured = capfd.readouterr()
            out = captured.out.strip()
            assert out == f"{mock_bin}/mpiexec -np 4 -xflag file executable --option"


def test_default(capfd):
    launch(["-n", "4", "-flag", "file", "executable", "--option"])
    captured = capfd.readouterr()
    out = captured.out.strip()
    assert out == f"{mock_bin}/mpiexec -n 4 -flag file executable --option"


def test_envar_mappings(capfd):
    with envmods(HPCC_LAUNCH_MAPPINGS="-spam:-ham,-eggs:-bacon"):
        launch(
            ["-n", "4", "-spam", "ham", "-eggs", "bacon", "executable", "--option"]
        )
        captured = capfd.readouterr()
        out = captured.out.strip()
        assert out == f"{mock_bin}/mpiexec -n 4 -ham ham -bacon bacon executable --option"


def test_mpmd():
    launcher = hpc_connect.get_launcher()
    argv = ["-n", "4", "-flag", "file", "ls", ":", "-n", "5", "ls", "-la"]
    cmd = launcher.prepare_command_line(argv)
    assert os.path.basename(cmd[0]) == "mpiexec"
    assert " ".join(cmd[1:]) == "-n 4 -flag file ls : -n 5 ls -la"


def test_srun_mpmd(tmpdir):
    with working_dir(str(tmpdir)):
        config = hpc_connect.config.Config()
        config.set("launch:exec", "srun")
        launcher = hpc_connect.get_launcher(config)
        argv = ["-n", "4", "ls", ":", "-n", "5", "ls", "-la"]
        cmd = launcher.prepare_command_line(argv)
        assert " ".join(cmd) == "srun -n9 --multi-prog launch-multi-prog.conf"
        with open("launch-multi-prog.conf") as fh:
            assert fh.read().strip() == "0-3 ls\n4-8 ls -la"


def test_mapped(tmpdir):
    with working_dir(str(tmpdir)):
        launcher = hpc_connect.get_launcher()
        launcher.config.set("launch:mappings", {"--x": "--y"})
        launcher.config.set("launch:numproc_flag", "-np")
        argv = ["--x", "4", "--x=5", "-n=7", "ls"]
        cmd = launcher.prepare_command_line(argv)
        assert os.path.basename(cmd[0]) == "mpiexec"
        assert " ".join(cmd[1:]) == "--y 4 --y=5 -np=7 ls"


def test_mapped_suppressed(tmpdir):
    with working_dir(str(tmpdir)):
        launcher = hpc_connect.get_launcher()
        launcher.config.set("launch:mappings", {"--x": "SUPPRESS"})
        launcher.config.set("launch:numproc_flag", "-np")
        argv = ["--x", "4", "--x=5", "-n=7", "ls"]
        cmd = launcher.prepare_command_line(argv)
        assert os.path.basename(cmd[0]) == "mpiexec"
        assert " ".join(cmd[1:]) == "-np=7 ls"


def test_count_procs(tmpdir):
    with working_dir(str(tmpdir)):
        from hpc_connect.launch.base import ArgumentParser
        parser = ArgumentParser(mappings={}, numproc_flag="-n")
        argv = ["-n", "4", "ls", ":", "-n=5", "ls"]
        args = parser.parse_args(argv)
        assert args.processes[0] == 4
        assert args.processes[1] == 5



# Copyright NTESS. See COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Any

from ..config import Config
from ..hookspec import hookimpl
from .base import HPCProcess
from .base import HPCSubmissionManager

try:
    import flux  # noqa: F401

except ImportError:

    class FluxSubmissionManager(HPCSubmissionManager):
        name = "flux"

        def __init__(self, config: Config) -> None:
            raise RuntimeError("FluxSubmissionManager requires the flux module be importable")

        def submit(
            self,
            name: str,
            args: list[str],
            scriptname: str | None = None,
            qtime: float | None = None,
            submit_flags: list[str] | None = None,
            variables: dict[str, str | None] | None = None,
            output: str | None = None,
            error: str | None = None,
            nodes: int | None = None,
            cpus: int | None = None,
            gpus: int | None = None,
            **kwargs: Any,
        ) -> HPCProcess:
            raise NotImplementedError

        @staticmethod
        def matches(name: str) -> bool:
            return name == "flux"

else:
    from .flux_api import FluxSubmissionManager  # type: ignore


@hookimpl
def hpc_connect_submission_manager(config) -> HPCSubmissionManager | None:
    if FluxSubmissionManager.matches(config.get("submit:backend")):
        return FluxSubmissionManager(config=config)
    return None



import argparse
import sys

from ..config import Config

description = "Show, get, and set config values"


def setup_parser(parser: argparse.ArgumentParser) -> None:
    parent = parser.add_subparsers(dest="subcommand")
    p = parent.add_parser("show")
    p.add_argument(
        "--scope",
        default=None,
        help="Add settings to this config scope [default: %(default)s]",
    )
    p = parent.add_parser("add")
    p.add_argument(
        "--scope",
        choices=("site", "global", "local"),
        default="local",
        help="Add settings to this config scope [default: %(default)s]",
    )
    p.add_argument(
        "add_config_paths",
        nargs="+",
        metavar="path",
        help="colon-separated path to config that should be set, e.g. 'config:default:true'",
    )


def execute(config: Config, args: argparse.Namespace) -> None:
    if args.subcommand == "show":
        config.dump(sys.stdout, scope=args.scope, default_flow_style=False)
    elif args.subcommand == "add":
        for path in args.add_config_paths:
            config.add(path, scope=args.scope)
